/**************
* Program:	DV Icy (Dreamcast VMU Icon Viewer / Converter)
* Author:	Sean Robinson
* License:	GNU Affero General Public License Version 3 (or later)
* Released:	2019-06-16
* Version:	0.6
**************/

#ifdef _WIN32
	#include <conio.h>
#endif

#include <fstream>
#include <iostream>
#include <string>
#include "SDL/SDL.h"
using namespace std;

void winPause()
{
#ifdef _WIN32
	cout << "Press any key to continue...";
	_getch();
	cin.clear();
#endif
}

void convertBWiconToArray(ifstream &vmuFile, bool bwIcon[])
{
	// Find black-and-clear icon starting offset
	vmuFile.seekg(16);
	vmuFile.seekg(vmuFile.get());

	// Translate from hexadecimal to pixels
	int pixelValues = vmuFile.get();
	int place = 0;
	for(int j = 0; j < 128; j++)
	{
		for(int i = 128; i > 0; i /= 2)
		{
			// Store icon pixels as trues and falses
			if(pixelValues >= i)
			{
				bwIcon[place] = true;
				pixelValues -= i;
			}
			else
			{
				bwIcon[place] = false;
			}
			place++;
		}
		pixelValues = vmuFile.get();
	}
}

void convertCLiconToArray(ifstream &vmuFile, int clIcon[])
{
	// Find color icon starting offset
	vmuFile.seekg(20);
	if(vmuFile.peek() == 0)
	{
		cout << "No color icon found in VMU file." << endl;
		// If no color icon, fill color data with 0's
		for(int k = 0; k < 1088; k++)
		{
			clIcon[k] = 0;
		}
		return;
	}
	vmuFile.seekg(vmuFile.get());

	// Translate from hexadecimal to 16 colors, 4 channels per color (16xGBAR=64)
	int pixelValues;
	for(int k = 0; k < 16; k++)
	{
		pixelValues = vmuFile.get();
		// Convert high-nybble to green channel, 0-15
		clIcon[k*4+0] = pixelValues / 16;
		// Convert low-nybble to blue channel, 0-15
		clIcon[k*4+1] = pixelValues % 16;
		pixelValues = vmuFile.get();
		// Convert high-nybble to alpha channel, 0-15
		clIcon[k*4+2] = pixelValues / 16;
		// Convert low-nybble to red channel, 0-15
		clIcon[k*4+3] = pixelValues % 16;
	}

	// Translate from hexadecimal to colors 0-15 for 32x32 pixels (1024)
	for(int k = 64; k < 1088; k += 2)
	{
		pixelValues = vmuFile.get();
		// Convert high-nybble to first color
		clIcon[k] = pixelValues / 16;
		// Convert low-nybble to second color
		clIcon[k+1] = pixelValues % 16;
	}
}

void displayBWIcon(bool bwIcon[])
{
	SDL_Surface *screen;

	/* Have a preference for 8-bit, but accept any depth */
	screen = SDL_SetVideoMode(320, 320, 8, SDL_SWSURFACE|SDL_ANYFORMAT);
	if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set 640x480x8 video mode: %s\n",
						SDL_GetError());
		exit(1);
	}
	printf("Set 640x480 at %d bits-per-pixel mode\n",
		   screen->format->BitsPerPixel);
	
	/* Code to set the screen to be yellow */

	Uint32 bgColor;

	/* Map the color yellow to this display (R=0xff, G=0xFF, B=0x00)
	   Note:  If the display is palettized, you must set the palette first.
	*/
	bgColor = SDL_MapRGB(screen->format, 0xff, 0x00, 0xff);
	Uint32 fgColor = SDL_MapRGB(screen->format, 0x00, 0x00, 0x00);

	/* Lock the screen for direct access to the pixels */
	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
			exit(1);
		}
	}

	SDL_FillRect(screen, NULL, bgColor);

	// Fill out the VMU icon's pattern
	SDL_Rect blockOfColor;
	int i = 0, k = 0;
	for(int j = 0; j < 1024; j++)
	{
		if(bwIcon[j])
		{
			blockOfColor.x = 10 * k;
			blockOfColor.y = 10 * i;
			blockOfColor.w = 10;
			blockOfColor.h = 10;
			SDL_FillRect(screen, &blockOfColor, fgColor);
		}
		k++;

		if((j + 1) % 32 == 0)
		{
			i++;
			k = 0;
		}
	}

	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
	/* Update just the part of the display that we've changed */
	SDL_Flip(screen);
	printf("Quiting SDL.\n");

	SDL_Delay(1000);

	/* Shutdown all subsystems */
	SDL_Quit();
}

void displayCLIcon(int clIcon[])
{
	SDL_Surface *screen;

	/* Have a preference for 8-bit, but accept any depth */
	screen = SDL_SetVideoMode(320, 320, 8, SDL_SWSURFACE|SDL_ANYFORMAT);
	if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set 640x480x8 video mode: %s\n",
						SDL_GetError());
		exit(1);
	}
	printf("Set 640x480 at %d bits-per-pixel mode\n",
		   screen->format->BitsPerPixel);

	Uint32 bgColor = SDL_MapRGB(screen->format, 0xff, 0x00, 0xff); // magenta

	/* Lock the screen for direct access to the pixels */
	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
			exit(1);
		}
	}

	SDL_FillRect(screen, NULL, bgColor);

	// Fill out the VMU icon's pattern
	SDL_Rect blockOfColor;
	blockOfColor.w = 10;
	blockOfColor.h = 10;
	int i = 0, k = 0;
	for(int j = 64; j < 1088; j++)
	{
		blockOfColor.x = 10 * k;
		blockOfColor.y = 10 * i;
		// Change color map from GBAR->RGBA
		Uint32 fgColor = SDL_MapRGBA(screen->format,
			clIcon[clIcon[j]*4+3]*17, clIcon[clIcon[j]*4+0]*17,
			clIcon[clIcon[j]*4+1]*17, clIcon[clIcon[j]*4+2]*17);
		SDL_FillRect(screen, &blockOfColor, fgColor);
		k++;

		if((j + 1) % 32 == 0)
		{
			i++;
			k = 0;
		}
	}

	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
	/* Update just the part of the display that we've changed */
	SDL_Flip(screen);
	printf("Quiting SDL.\n");

	SDL_Delay(1000);

	/* Shutdown all subsystems */
	SDL_Quit();
}

void printBWIcon(bool bwIcon[])
{
	// Display icon as X's and white spaces
	for(int j = 0; j < 1024; j++)
	{
		if(bwIcon[j])
		{
			cout << "X";
		}
		else
		{
			cout << " ";
		}

		if((j + 1) % 32 == 0)
		{
			cout << endl;
		}
	}
}

void printHelpInfo(char helpType = 'h')
{
	const float version = 0.3;
	// Print version info
	cout << "DV Icy (Dreamcast VMU Icon Converter)" << endl;
	cout << "Version " << version << endl << endl;

	// Print help and program options
	if(helpType == 'h')
	{

		cout << "-b, --printb" << endl;
		cout << "\t" << "print the black-and-clear icon to standard output" << endl << endl;

		cout << "-B, --displayb" << endl;
		cout << "\t" << "display the black-and-clear icon on screen" << endl << endl;

		cout << "-c, --printc" << endl;
		cout << "\t" << "print the color icon (in black) to standard output" << endl << endl;

		cout << "-C, --displayc" << endl;
		cout << "\t" << "display the color icon on screen" << endl << endl;

		cout << "-h, --help" << endl;
		cout << "\t" << "display this help and exit" << endl << endl;

		cout << "-v, --version" << endl;
		cout << "\t" << "output version information and exit" << endl << endl;
	}
	//winPause();
}

int main(int argc, char *argv[])
{
	// Try to reclaim standard output and standard error from SDL (on Windows)
	#ifdef _WIN32
		freopen( "CON", "w", stdout);
		freopen( "CON", "w", stderr);
	#endif

	bool help = false, version = false;
	bool displayBW = false, displayCL = false;
	bool printBW = false, printCL = false;
	char *vmuFileName = "";

	// Parse command line arguments
	for(int i = 1; i < argc; i++)
	{
		if(strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
			help = true;
		else if(strcmp(argv[i], "--version") == 0 || strcmp(argv[i], "-v") == 0)
			version = true;
		else if(strcmp(argv[i], "--displayb") == 0 || strcmp(argv[i], "-B") == 0)
			displayBW = true;
		else if(strcmp(argv[i], "--displayc") == 0 || strcmp(argv[i], "-C") == 0)
			displayCL = true;
		else if(strcmp(argv[i], "--printb") == 0 || strcmp(argv[i], "-b") == 0)
			printBW = true;
		else if(strcmp(argv[i], "--printc") == 0 || strcmp(argv[i], "-c") == 0)
			printCL = true;
		else
			vmuFileName = argv[i];
	}

	// Print help or version info and exit
	if(help)
	{
		printHelpInfo('h');
		return 0;
	}
	else if(version)
	{
		printHelpInfo('v');
		return 0;
	}

	// If there is no file name, print help screen
	if(vmuFileName == "")
	{
		printHelpInfo('h');
		return 1;
	}

	// Read VMU file
	ifstream vmuFile(vmuFileName, ifstream::binary);
	if(vmuFile.good())
		cout << "Successfully read \"" << vmuFileName << "\"." << endl << endl;
	else
	{
		cerr << "Invalid file name or file not found: \"" << vmuFileName << "\".";
		cerr << endl << endl;

		// Pause to allow user to look at output (on Windows)
		//winPause();
		return 1;
	}

	// Arrays for the icon name, the black-and-clear icon, and the color icon
	char iconName [17];
	bool bwIcon [1024];
	int clIcon [1088];

	// Convert icon description to character array
	for(int i = 0; i < 16; i++)
		iconName[i] = (char) vmuFile.get();
	iconName[16] = '\0';

	// Convert black-and-clear icon to boolean array
	if(displayBW || printBW)
	{
		convertBWiconToArray(vmuFile, bwIcon);
	}

	// Convert black-and-clear icon to boolean array
	if(displayCL || printCL)
	{
		convertCLiconToArray(vmuFile, clIcon);
	}

	// Close VMU file
	vmuFile.close();

	// Display comment/description and icon on screen display
	if(displayBW || displayCL)
	{
		// Initialize SDL
		if(SDL_Init(SDL_INIT_VIDEO) < 0)
		{
			cerr << "Unable to initialize SDL: " << SDL_GetError() << endl;
			return 1;
		}
		atexit(SDL_Quit);

		if(displayBW)
			displayBWIcon(bwIcon);
		else
			displayCLIcon(clIcon);
	}

	// Print comment/description and icon to terminal/console
	if(printBW || printCL)
	{
		cout << "Icon description/name: ";
		cout << iconName << endl << endl;

		// Print the black-and-clear icon to standard output
		if(printBW)
			cout << "Black-and-clear icon: ";
			cout << endl;
			printBWIcon(bwIcon);
		cout << endl;

		// Pause to allow user to look at output (on Windows)
		//winPause();
	}

	return 0;
}
